'''
bwmember -- Easy to build type-safe propreties

>>> class MyClass(BWObject):
...     x = member(int, str)
...
>>> c = MyClass(x=5)
>>> c.x
5
>>> c.x = 7
>>> c.x
7
>>> c.x = 'Hello'
>>> c.x
'Hello'
>>> c.x = None
Traceback (most recent call last):
    ...
TypeError: None ('x') must be one of: (int, str)
'''

from __version__ import *
from bwobject import BWObject
from bwmethod import after_super
from bwcached import cached

class BWMemberProperty(property):
    pass

class BWMember(BWObject):
    ro = False

    def __init__(self, *isa, **_kw):
        self.isa = isa
        self.init(**_kw)

    def init(self, ro=None):
        if ro is not None:
            self.ro = ro

    def __bindclass__(self, cls, name):
        p = BWMemberProperty(self.get_reader(cls, name),
                             self.get_writer(cls, name),
                             self.get_deleter(cls, name))
        p.__initobj__ = self.__initobj__
        return p

    def typecheck(self, obj):
        for check in self.isa:
            if isinstance(obj, check):
                return True
            elif check is None and obj is None:
                return True
        return False

    def __initobj__(self, obj, name, value):
        if not self.typecheck(value):
            raise TypeError('%r (%r) must be one of: (%s)'
                            % (value, name,
                               ', '.join(t.__name__ for t in self.isa)))
        obj.__dict__[name,] = value

    def get_reader(self, cls, name):
        return lambda o: o.__dict__[name,]

    def get_writer(self, cls, name):
        if self.ro:
            return None
        else:
            return lambda o, v: self.__initobj__(o, name, v)

    def get_deleter(self, cls, name):
        if self.ro:
            return None
        else:
            return None

def member(*_args, **_kw):
    return BWMember(*_args, **_kw)

def ro_member(*_args, **_kw):
    _kw.setdefault('ro', True)
    return BWMember(*_args, **_kw)




'''
__changelog__ -- manage the changelog.
'''

class Version(tuple):
    '''
    >>> Version(0.1)
    0.1
    >>> Version('0.1.2')
    0.1.2
    >>> Version('0.1.2') > Version('0.1')
    True
    '''

    def __new__(cls, src, *changes):
        if src is None or isinstance(src, Version):
            return src
        elif isinstance(src, basestring):
            return Version(src.split('.'), *changes)
        elif isinstance(src, (int, float)):
            return Version(str(src), *changes)
        else:
            obj = super(Version, cls).__new__(cls, src)
            obj.changes = changes
            if len(changes) == 0:
                obj.label = None
            elif len(changes) == 1:
                obj.label = changes[0]
            else:
                obj.label = changes
            return obj

    def __add__(self, other):
        if isinstance(other, basestring):
            return str(self) + other
        else:
            return super(Version, self).__add__(other)

    def __radd__(self, other):
        if isinstance(other, basestring):
            return other + str(self)
        else:
            return other.__add__(self)

    def __str__(self):
        return '.'.join(self)
    __repr__ = __str__

class ChangeLog(tuple):
    '''
    >>> ChangeLog(
    ...     Version(0.1, 'Version 0.1'),
    ...     Version(0.2, 'Change A in 0.2', 'Change B in 0.2'),
    ... )
    ...
    0.2
        * Change A in 0.2
        * Change B in 0.2
    <BLANKLINE>
    0.1
        * Version 0.1
    '''

    def __new__(cls, *items):
        return super(ChangeLog, cls).__new__(cls, items)

    def __str__(self):
        return '\n\n'.join(
            '%s\n    * %s' % (
                version,
                '<No log provided>' if version.label is None
                    else version.label if len(version.changes) == 1
                    else '\n    * '.join(map(str, version.changes))
                ) for version in reversed(sorted(self)))

    __repr__ = __str__

CHANGELOG = ChangeLog(
    Version('0.1', 'Initial pre-release at PyOhio'),
    Version('0.2.5',
        'Added throwables',
        'Added changelog file',
        'Modified version to be more shell friendly',
        'Made __version__ more pydoc compatible',
        ),
    Version('0.2.6',
        'Added throw() and catch() as simple functions',
        )
)

if __name__ == '__main__':
    print CHANGELOG

